@@ -4,7 +4,6 @@
 
 source $TOP/tests/common
 
-check_prereq btrfs-show-super
 check_prereq mkfs.btrfs
 check_prereq btrfs
 
@@ -38,7 +37,7 @@ cleanup_devices()
 
 test_get_info()
 {
-	run_check $TOP/btrfs-show-super $dev1
+	run_check $TOP/btrfs inspect-internal dump-super $dev1
 	run_check $SUDO_HELPER $TOP/btrfs check $dev1
 	run_check $SUDO_HELPER mount $dev1 $TEST_MNT
 	run_check $TOP/btrfs filesystem df $TEST_MNT
@@ -70,6 +69,8 @@ test_mkfs_single
 test_mkfs_single  -d  single  -m  single
 test_mkfs_single  -d  single  -m  single  --mixed
 test_mkfs_single  -d  single  -m  dup
+test_mkfs_single  -d  dup     -m  single
+test_mkfs_single  -d  dup     -m  dup
 test_mkfs_single  -d  dup     -m  dup     --mixed
 
 test_mkfs_multi
@@ -85,5 +86,7 @@ test_mkfs_multi   -d  raid5   -m  raid5
 test_mkfs_multi   -d  raid5   -m  raid5   --mixed
 test_mkfs_multi   -d  raid6   -m  raid6
 test_mkfs_multi   -d  raid6   -m  raid6   --mixed
+test_mkfs_multi   -d  dup     -m  dup
+test_mkfs_multi   -d  dup     -m  dup     --mixed
 
 cleanup_devices
